# creates menus in the Mac OS X menu bar

# File and Edit menus are standard
proc makeStandardMacMenus {win {destroywin 0} {ndw -2}} {
    
    # make a menu bar
    menu $win.mbar -tearoff 0
    if {$win == ""} {
        . configure -menu .mbar
    } else {
        $win configure -menu $win.mbar
    }
    
    $win.mbar add cascade -label File -menu $win.mbar.file
    menu $win.mbar.file -tearoff 0
    if {$win == ""} {
        $win.mbar.file add command -label "Open\u2026"     -command "openFile .input"     -accel "Command-O"
        $win.mbar.file add command -label "Close"          -accel "Command-W"       -state disabled
    } else {
        if {$win == ".output"} {
            $win.mbar.file add command -label "Open\u2026"     -command "openFile .output"     -accel "Command-O"
        } else {
            $win.mbar.file add command -label "Open\u2026"     -command "openFile .input"   -accel "Command-O"
        }
        if {$win == ".input"} {
            set closeAction "closeInput"
        } elseif {$win == ".output"} {
            set closeAction ".output.fr1.text delete 1.0 end ; wm withdraw $win"
        } elseif {$ndw > -1} {
            # knot drawing window
            set closeAction [list knotPictureDelete $ndw]
        } elseif {$win == ".canonicalPicture"} {
            set closeAction [list canonicalPictureDelete]
        } elseif {$destroywin} {
            set closeAction "destroy $win"
        } else {
            set closeAction "wm withdraw $win"
        }
        $win.mbar.file add command -label "Close"          -command $closeAction        -accel "Command-W"
    }
    $win.mbar.file add separator
    if {$win == ".input" || $win == ".output"} {
        set saveAction "saveFile $win"
        $win.mbar.file add command -label "Save"           -command $saveAction         -accel "Command-S"
        $win.mbar.file add command -label "Save As\u2026"  -command "saveAs $win"       -accel "Command-Shift-S"
    } else {
        if {$ndw > -1} {
            # knot drawing window
            set saveAction "knotPictureSavePs $ndw"
            $win.mbar.file add command -label "Save PostScript\u2026" -command $saveAction \
                -accel "Command-S"
        } elseif {$win == ".canonicalPicture"} {
            set saveAction "canonicalSavePs"
            $win.mbar.file add command -label "Save PostScript\u2026" -command $saveAction \
                -accel "Command-S"
        } else {
            $win.mbar.file add command -label "Save"       -accel "Command-S"       -state disabled
        }
        $win.mbar.file add command -label "Save As\u2026"  -accel "Command-Shift-S" -state disabled
    }
    $win.mbar.file add separator
    $win.mbar.file add command -label "LinkSmith"      -command "showLinkSmith x"   -accel "Command-K"
    $win.mbar.file add command -label "Enter Braid"    -command "enterBraid"        -accel "Command-Shift-B"
    $win.mbar.file add command -label "Browse"         -command "showTableChoice x" -accel "Command-B"
    
    bind $win <Command-o> {openFile .input}
    if {[info exists closeAction]} {
        bind $win <Command-w> $closeAction
    }
    bind $win <Command-k> {showLinkSmith x}
    if {[info exists saveAction]} {
        bind $win <Command-s> $saveAction
    }
    bind $win <Command-S> {saveAs .input}
    bind $win <Command-B> {enterBraid}
    bind $win <Command-b> {showTableChoice x}
    bind $win <Command-q> {tidyExit}

    $win.mbar add cascade -menu $win.mbar.edit -label Edit
    menu $win.mbar.edit -tearoff 0
    $win.mbar.edit add command -label Cut -accel Command-X -command { event generate [focus] <<Cut>> }
    $win.mbar.edit add command -label Copy -accel Command-C -command { event generate [focus] <<Copy>> }
    $win.mbar.edit add command -label Paste -accel Command-V -command { event generate [focus] <<Paste>> }
    $win.mbar.edit add command -label Clear -command { event generate [focus] <<Clear>> }
    bind $win <Command-X> { event generate [focus] <<Cut>> }
    bind $win <Command-C> { event generate [focus] <<Copy>> }
    bind $win <Command-V> { event generate [focus] <<Paste>> }

    $win.mbar add cascade -menu $win.mbar.help
    menu $win.mbar.help -tearoff 0
    $win.mbar.help add command -label "Knotscape Help" -command {showHelp toc} -accel "Command-?"
    bind $win <Command-?> { showHelp toc }
    
    $win.mbar add cascade -menu $win.mbar.apple
    menu $win.mbar.apple -tearoff 0
    $win.mbar.apple add command -label "About Knotscape" -command { aboutKnotscape }
}

# add all appropriate menus to all appropriate windows
proc makeMacMenus {} {

    # make default menu for the top level
    makeStandardMacMenus ""
    
    # .input needs some an extra menu, Action
    makeStandardMacMenus .input
    .input.mbar add cascade -label Action -menu .input.mbar.action
    menu .input.mbar.action -tearoff 0
    foreach item {{"Locate in Table"         "locateInTable"}
                  {"Draw Knot"               "drawKnot"}
                  {"Make Postscript"         "psKnot"}
                  {"Compute Polynomials"     "showPolyChoice x"}
                  {"Find Homomorphisms"      "showHomChoice x"}
                  {"Determinant & Signature" "detSigCompute"}} {
        .input.mbar.action add command -label [lindex $item 0]\
                                       -command [lindex $item 1]
    } 
    .input.mbar.action add cascade -label "Hyperbolic Invariants"\
                                   -menu .input.mbar.action.hyperbolic
    menu .input.mbar.action.hyperbolic -tearoff 0
    foreach item {{"Volume"                        "hypInvariants 1"}
                  {"Symmetry Group"                "hypInvariants 2"}
                  {"Maximal Cusp Neighborhood"     "hypInvariants 3"}
                  {"Detailed Symmetry Information" "symmetryInfo"}
                  {"Canonical Cell Decomposition"  "canonicalPicture"}
                  {"Save Triangulation"            "saveKnotTriangulation"}} {
        .input.mbar.action.hyperbolic add command -label [lindex $item 0]\
                                                  -command [lindex $item 1]
    }
    
    # most things need standard menus
    makeStandardMacMenus .output
    makeStandardMacMenus .polynomials
    makeStandardMacMenus .tableChoice
    makeStandardMacMenus .homChoice
    
    # LinkSmith has entirely different menus
    # we use 'macbar' here because 'mbar' is the non-Mac "menu bar"
    menu .linksmith.macbar -tearoff 0
    .linksmith configure -menu .linksmith.macbar
    
    .linksmith.macbar add cascade -label "File" -menu .linksmith.macbar.file
    menu .linksmith.macbar.file -tearoff 0
    .linksmith.macbar.file add command -label "Open\u2026" -command openDiagram
    .linksmith.macbar.file add separator
    .linksmith.macbar.file add command -label "Save" -command saveDiagram
    .linksmith.macbar.file add command -label "Save As\u2026" -command saveAsDiagram
    .linksmith.macbar.file add separator
    .linksmith.macbar.file add command -label "Submit to Knotscape" -command submitToKnotscape -accel "Command-T"
    .linksmith.macbar.file add separator
    .linksmith.macbar.file add cascade -label "Close"\
                                       -menu .linksmith.macbar.file.closeOptions
    .linksmith.macbar.file add cascade -label "Quit LinkSmith"\
                                       -menu .linksmith.macbar.file.quitOptions
    
    menu .linksmith.macbar.file.closeOptions -tearoff 0
    menu .linksmith.macbar.file.quitOptions -tearoff 0
      
    .linksmith.macbar.file.closeOptions add command -label "Save diagram"\
                                                    -command {closeDiagram noquit save}
    .linksmith.macbar.file.closeOptions add command -label "Don't save"\
                                                    -command {closeDiagram noquit nosave}      

    .linksmith.macbar.file.quitOptions add command -label "Save diagram"\
                                                   -command {closeDiagram quit save}
    .linksmith.macbar.file.quitOptions add command -label "Don't save"\
                                                   -command {closeDiagram quit nosave}

    bind .linksmith <Command-t> submitToKnotscape

    .linksmith.macbar add cascade -label "Edit" -menu .linksmith.macbar.edit
    menu .linksmith.macbar.edit -tearoff 0
    .linksmith.macbar.edit add command -label "Delete" -command {deleteselectedEdges "yes"}
    .linksmith.macbar.edit add command -label "Clear" -command startOver
    
    bind .linksmith <Key-BackSpace> {deleteselectedEdges "yes"}
    bind .linksmith <Key-Delete> {deleteselectedEdges "yes"}

    .linksmith.macbar add cascade -label "Action" -menu .linksmith.macbar.action
    menu .linksmith.macbar.action -tearoff 0
    .linksmith.macbar.action add command -label "Alternate" -command alternate
    .linksmith.macbar.action add command -label "Reflect" -command reflect
    .linksmith.macbar.action add command -label "Reverse" -command reverseSelection
    .linksmith.macbar.action add command -label "Reverse All" -command reverseAll

    .linksmith.macbar add cascade -label "Export" -menu .linksmith.macbar.export
    menu .linksmith.macbar.export -tearoff 0
    .linksmith.macbar.export add command -label "Save Millett Code" -command saveMillettCode
    .linksmith.macbar.export add command -label "Save Triangulation" -command saveTriangulation

    .linksmith.macbar add cascade -menu .linksmith.macbar.help
    menu .linksmith.macbar.help -tearoff 0
    .linksmith.macbar.help add command -label "LinkSmith Help" -command {showHelp linksmith} -accel "Command-?"
    
    bind .linksmith <Command-?> {showHelp linksmith}
}

proc aboutKnotscape {} {
    toplevel .aboutKnotscape
    wm title .aboutKnotscape "About Knotscape"
        
    message .aboutKnotscape.app -width 400 -text "Knotscape" -justify center
    message .aboutKnotscape.details -text "version 1.0.2\n\nby Morwen Thistlewaite\nand\nJim Hoste" -justify center
    button .aboutKnotscape.ok -text "OK" -command "destroy .aboutKnotscape"
    pack .aboutKnotscape.app -expand 1 -padx 3m -pady 3m
    pack .aboutKnotscape.details -expand 1 -padx 3m -pady 3m
    pack .aboutKnotscape.ok -pady 3m
    
    set x [expr [winfo screenwidth .aboutKnotscape]/2 - [winfo reqwidth .aboutKnotscape]/2 - \
        [winfo vrootx [winfo parent .aboutKnotscape]]]
    set y [expr [winfo screenheight .aboutKnotscape]/2 - [winfo reqheight .aboutKnotscape]/2 - \
        [winfo vrooty [winfo parent .aboutKnotscape]]]
    wm geometry .aboutKnotscape +$x+$y
    
    wm protocol .aboutKnotscape WM_DELETE_WINDOW { grab release .aboutKnotscape; destroy .aboutKnotscape }
    
    wm resizable .aboutKnotscape 0 0
    raise .aboutKnotscape
    tkwait window .aboutKnotscape
}